from fastapi.testclient import TestClient
import sys
import polars as pl
import pandas as pd
import pytest


sys.path.append(".")

engines = ("duckdb", "polars")
# engines = ("duckdb",)


def test_no_authentication(client_no_auth: TestClient):
    response = client_no_auth.get("/")
    assert response.status_code == 401
    assert response.json() == {"detail": "Not authenticated"}


def test_authentication(client: TestClient):
    response = client.get("/")
    assert response.status_code == 200


@pytest.mark.parametrize("engine", engines)
def test_fruits_limit_1(engine, client: TestClient):
    for _ in range(2):
        response = client.get(
            f"/api/v1/test/fruits?limit=1&format=json&%24engine={engine}"
        )
        assert (
            response.headers["Content-Type"].replace(" ", "")
            == "application/json;charset=utf-8"
        )
        assert response.status_code == 200
        assert response.json() == [
            {
                "A": 1,
                "fruits": "banana",
                "B": 5,
                "cars": "beetle",
            }
        ]


def test_fruits_sort_asc(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_sortby_asc?limit=1&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 5,
                    "fruits": "banana",
                    "B": 1,
                    "cars": "beetle",
                }
            ]


def test_fruits_sort_desc(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_sortby_desc?limit=1&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 9,
                    "fruits": "ananas",
                    "B": 9,
                    "cars": "fiat",
                }
            ]


def test_fruits_offset_1(client: TestClient):
    for _ in range(2):
        for e in engines:  # polars does not support offset
            response = client.get(
                f"/api/v1/test/fruits?limit=1&&offset=1&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                }
            ]


def test_data_limit(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fake_delta?limit=1000&&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert len(response.json()) == 1000


def test_data_filter(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1&format=json&cars=audi&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                }
            ]
            response = client.get(
                f"/api/v1/test/fruits?limit=1&format=json&fruits=ananas&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 9,
                    "fruits": "ananas",
                    "B": 9,
                    "cars": "fiat",
                }
            ]


def test_data_csv(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1&format=csv&cars=audi&%24engine={e}",
            )
            assert response.status_code == 200
            txt = response.text
            import csv

            reader = csv.DictReader(txt.splitlines())
            line1 = reader.__next__()
            assert line1 == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}


def test_data_scsv(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1&format=scsv&cars=audi&%24engine={e}",
            )
            assert response.status_code == 200
            txt = response.text
            import csv

            reader = csv.DictReader(txt.splitlines(), delimiter=";")
            line1 = reader.__next__()
            assert line1 == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}


def test_data_tsv(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1&format=csv&$csv_separator=\\t&cars=audi&%24engine={e}",
            )
            assert response.status_code == 200
            txt = response.text
            import csv

            reader = csv.DictReader(txt.splitlines(), delimiter="\t")
            line1 = reader.__next__()
            assert line1 == {"A": "2", "fruits": "banana", "B": "4", "cars": "audi"}


def test_data_minus_1(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fake_delta?limit=-1&format=json&%24engine={e}"
            )
            assert response.status_code == 200
            assert len(response.json()) == 100_011


def test_data_minus_1_csv(client: TestClient):
    for _ in range(2):
        from io import StringIO

        for f in ("csv", "scsv"):
            separator = ";" if f == "scsv" else ","
            for e in engines:
                response = client.get(
                    f"/api/v1/test/fake_delta?limit=-1&format={f}&%24engine={e}",
                )
                assert response.status_code == 200
                csv_file = StringIO(response.text)
                assert len(pd.read_csv(csv_file, sep=separator)) == 100_011


def test_data_default_limit(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(f"/api/v1/test/fake_delta?format=json&%24engine={e}")
            assert response.status_code == 200
            assert len(response.json()) == 100


def test_get_parquet_data(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1000&format=parquet&%24engine={e}",
            )
            assert response.status_code == 200
            from io import BytesIO

            buf = BytesIO(response.content)
            df = pd.read_parquet(buf)
            df2 = pl.read_delta("tests/data/delta/fruits").to_pandas()
            assert df.equals(df2)


def test_get_arrow_data(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1000&format=arrow&%24engine={e}",
            )
            assert response.status_code == 200
            from io import BytesIO

            buf = BytesIO(response.content)
            df = pl.read_ipc(buf).to_pandas()
            df2 = pl.read_delta("tests/data/delta/fruits").to_pandas()
            assert df.equals(df2)


def test_get_ndjson_data(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1000&format=ndjson&%24engine={e}",
            )
            assert response.status_code == 200
            from io import BytesIO

            buf = BytesIO(response.content)
            df = pl.read_ndjson(buf).to_pandas()
            df2 = pl.read_delta("tests/data/delta/fruits").to_pandas()
            assert df.equals(df2)


def test_get_excel_data(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits?limit=1000&format=xlsx&%24engine={e}",
            )
            assert response.status_code == 200
            from io import BytesIO

            buf = BytesIO(response.content)
            df = pl.read_excel(buf).to_pandas()
            df2 = pl.read_delta("tests/data/delta/fruits").to_pandas()
            assert df.equals(df2)


def test_fruits_in(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"fruits_in": ["banana", "ananas"]},
            )
            assert response.status_code == 200
            assert len(response.json()) == 4

            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"fruits_in": ["apple", "ananas"]},
            )
            assert response.status_code == 200
            assert len(response.json()) == 4


def test_fruits_in_zero(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"A_in": [0, 9]},
            )
            assert response.status_code == 200
            assert len(response.json()) == 2


def test_fruits_combi(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"pk": [{"cars": "audi"}]},
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                }
            ]


def test_fruits_combi_int(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"combiint": [{"A": 0, "cars": "lamborghini", "B": 5}]},
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 0,
                    "fruits": "apple",
                    "B": 5,
                    "cars": "lamborghini",
                }
            ]

            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"combiint": [{"A": 1, "cars": "beetle", "B": 5}]},
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 1,
                    "fruits": "banana",
                    "B": 5,
                    "cars": "beetle",
                }
            ]


def test_fruits_combi_different_name(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={"combi": [{"cars": "audi", "fruits": "banana"}]},
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                }
            ]


def test_fruits_combi_multi(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.post(
                "/api/v1/test/fruits?limit=1000",
                json={
                    "pk": [
                        {"cars": "audi", "fruits": "banana"},
                        {"cars": "fiat", "fruits": "ananas"},
                    ]
                },
            )
            assert response.status_code == 200
            assert len(response.json()) == 2
            assert response.json() == [
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                },
                {
                    "A": 9,
                    "fruits": "ananas",
                    "B": 9,
                    "cars": "fiat",
                },
            ]


def test_fruits_select(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_select?limit=1&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [{"fruits_new": "banana"}]


def test_fake_parquet(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fake_parquet?limit=10&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert len(response.json()) == 10


def test_fake_parquet_ns(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fake_parquet_ns?limit=10&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert len(response.json()) == 10


def test_fake_csv(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_csv?limit=3&format=json&%24engine={e}"
            )
            assert response.status_code == 200
            assert len(response.json()) == 3


def test_json(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_json?limit=3&format=json&%24engine={e}"
            )
            assert response.status_code == 200
            assert len(response.json()) == 3


def test_ndjson(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_ndjson?limit=3&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert len(response.json()) == 3


def test_fake_arrow(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fake_arrow?limit=10&format=json&%24engine={e}"
            )
            assert response.status_code == 200
            assert len(response.json()) == 10


def test_all_metadata(client: TestClient):
    response = client.get("/metadata")
    assert response.status_code == 200
    jsd = response.json()
    for item in jsd:
        for e in engines:
            name = item["name"]
            tag = item["tag"]
            route = item["route"]
            meta_detail_route = route + f"/metadata_detail?%24engine={e}"
            print(meta_detail_route)
            response = client.get(meta_detail_route)
            if name not in ["not_existing", "not_existing2"]:
                assert name + "_" + str(response.status_code) == name + "_200"
            else:
                assert name + "_" + str(response.status_code) == name + "_404"

    response = client.get("/api/v1/test/fake_arrow/metadata_detail")
    assert response.status_code == 200
    jsd = response.json()
    assert len(jsd["parameters"]) == 2


def test_metadata_stringify(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/complexer/complex_fruits/metadata_detail?jsonify_complex=True&%24engine={e}",
        )
        assert response.status_code == 200
        assert response.json()["modified_date"] is not None


def test_metadata_no_str_length(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/complexer/complex_fruits/metadata_detail?include_str_lengths=False&%24engine={e}",
        )
        assert response.status_code == 200
        assert response.json()["modified_date"] is not None


def test_metadata_no_hidden(client: TestClient):
    response = client.get("/api/v1/test/fruits_partition/metadata_detail")
    assert response.status_code == 200
    jsd = response.json()
    prm_names = [p["name"] for p in jsd["parameters"]]
    assert "cars_md5_prefix_2" not in prm_names
    field_names = [f["name"] for f in jsd["data_schema"]]
    assert "cars_md5_prefix_2" not in field_names


def test_auth_metadata(client_no_auth: TestClient):
    response = client_no_auth.get("/metadata")
    assert response.status_code == 401

    response = client_no_auth.get("/api/v1/test/fake_arrow/metadata_detail")
    assert response.status_code == 401


def test_fruits_nested(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/fruits_nested?limit=2&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            assert response.json() == [
                {
                    "A": 1,
                    "fruits": "banana",
                    "B": 5,
                    "cars": "beetle",
                    "nested": {"fruits": "banana", "cars": "beetle"},
                },
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                    "nested": {"fruits": "banana", "cars": "audi"},
                },
            ] or response.json() == [
                {
                    "A": 1,
                    "fruits": "banana",
                    "B": 5,
                    "cars": "beetle",
                    "nested": {"cars": "beetle", "fruits": "banana"},
                },
                {
                    "A": 2,
                    "fruits": "banana",
                    "B": 4,
                    "cars": "audi",
                    "nested": {"cars": "audi", "fruits": "banana"},
                },
            ]


def test_sortby(client: TestClient):
    for _ in range(2):
        for e in engines:
            response = client.get(
                f"/api/v1/test/search_sample?limit=5&format=json&%24engine={e}",
            )
            assert response.status_code == 200
            jsd = response.json()
            assert len(jsd) == 5

            assert jsd[0]["randomdata"] <= jsd[1]["randomdata"]
            assert jsd[1]["randomdata"] <= jsd[2]["randomdata"]
            assert jsd[2]["randomdata"] <= jsd[3]["randomdata"]
            assert jsd[3]["randomdata"] <= jsd[4]["randomdata"]



import sys

sys.path.append(".")

from fastapi.testclient import TestClient
from .utils import get_app, get_auth

auth = get_auth()

engines = ("duckdb", "polars")


def test_openid():
    for engine in engines:
        client = TestClient(get_app(default_engine=engine))
        response = client.get("/openapi.json", auth=auth)
        assert response.status_code == 200
        jsd = response.json()
        assert isinstance(jsd, dict)
        paths = jsd["paths"]
        schema = jsd["components"]["schemas"]
        ## TODO : Add more tests for different endpoints
        assert "test_fake_delta_partition" in schema.keys()
        assert "test_fake_polars_postParameter" in schema.keys()
        assert "abc" in schema["test_fake_polars_postParameter"]["properties"]
        assert "/api/v1/startest/fruits" in paths
        assert isinstance(paths["/api/v1/startest/fruits"], dict)
        assert len(paths["/api/v1/startest/fruits"].keys()) == 1
        assert "get" in paths["/api/v1/startest/fruits"].keys()

        assert isinstance(paths["/api/v1/startest/fruits_partition"], dict)
        assert len(paths["/api/v1/startest/fruits_partition"].keys()) == 1
        assert "post" in paths["/api/v1/startest/fruits_partition"].keys()



from typing import Literal, Tuple, cast

from fastapi import APIRouter
from bmsdna.lakeapi.context import ExecutionContextManager

from bmsdna.lakeapi.core.config import BasicConfig, Configs
from bmsdna.lakeapi.core.log import get_logger


logger = get_logger(__name__)

all_lake_api_routers: list[Tuple[BasicConfig, Configs]] = []


def init_routes(configs: Configs, basic_config: BasicConfig):
    from bmsdna.lakeapi.endpoint.endpoint import (
        get_response_model,
        create_config_endpoint,
    )
    from bmsdna.lakeapi.endpoint.detail_endpoint import create_detailed_meta_endpoint
    from bmsdna.lakeapi.endpoint.sql_endpoint import create_sql_endpoint
    from bmsdna.lakeapi.core.schema_cache import get_schema_cached

    all_lake_api_routers.append((basic_config, configs))
    router = APIRouter()
    metadata = []
    with ExecutionContextManager(
        basic_config.default_engine,
        basic_config.default_chunk_size,
    ) as mgr:
        for config in configs:
            methods = (
                cast(list[Literal["get", "post"]], [config.api_method])
                if isinstance(config.api_method, str)
                else config.api_method
            )
            try:
                from bmsdna.lakeapi.core.datasource import Datasource

                assert config.datasource is not None
                with Datasource(
                    config.version_str,
                    config.tag,
                    config.name,
                    config=config.datasource,
                    sql_context=mgr.get_context(config.engine),
                    basic_config=basic_config,
                    accounts=configs.accounts,
                ) as realdataframe:
                    try:
                        schema = get_schema_cached(
                            basic_config,
                            realdataframe,
                            config.datasource.get_unique_hash(),
                        )
                        if schema is None:
                            logger.warning(
                                f"Could not get response type for {config.route}. Path does not exist:{realdataframe}"
                            )
                    except Exception as err:
                        logger.warning(
                            f"Could not get schema for {config.route}. Error:{err}",
                            exc_info=err,
                        )
                        schema = None

                metadata.append(
                    {
                        "name": config.name,
                        "tag": config.tag,
                        "route": config.route,
                        "methods": methods,
                        "file_type": config.datasource.file_type,
                        "uri": config.datasource.uri,
                        "version": config.version,
                        "schema": {n: str(schema.field(n).type) for n in schema.names}
                        if schema
                        else None,
                    }
                )

            except Exception as err:
                import traceback

                print(traceback.format_exc())
                logger.warning(
                    f"Could not get response type for f{config.route}. Error:{err}"
                )
                schema = None

            response_model = (
                get_response_model(
                    config=config,
                    schema=schema,
                    basic_config=basic_config,
                )
                if schema is not None
                else None
            )
            create_detailed_meta_endpoint(
                schema=schema,
                config=config,
                configs=configs,
                router=router,
                basic_config=basic_config,
            )
            for am in methods:
                create_config_endpoint(
                    apimethod=am,
                    config=config,
                    router=router,
                    response_model=response_model,
                    schema=schema,
                    basic_config=basic_config,
                    configs=configs,
                )

        @router.get(
            "/metadata",
            name="metadata",
        )
        async def get_metadata():
            return metadata

        if basic_config.enable_sql_endpoint:
            create_sql_endpoint(
                router=router,
                basic_config=basic_config,
                configs=configs,
            )

        return router



import polars as pl

sql_cont = pl.SQLContext()
sql_cont.register("test_fake_arrow", pl.scan_ipc("tests/data/arrow/fake.arrow"))
print(sql_cont.execute("SELECT * FROM test_fake_arrow s LIMIT 1").collect())



import sys

sys.path.append(".")

engines = ["duckdb"]


from fastapi.testclient import TestClient


def test_search(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/test/search_sample?limit=5&format=json&%24engine={e}&search=Karen%20example"
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) >= 3
        assert len(jsd) <= 5
        assert jsd[0]["search_score"] is not None
        assert jsd[1]["search_score"] is not None
        assert jsd[2]["search_score"] is not None

        assert jsd[0]["search_score"] >= jsd[1]["search_score"]
        assert jsd[1]["search_score"] >= jsd[2]["search_score"]
        for item in jsd:
            assert (
                "karen"
                in (item["email"] + " " + item["name"] + " " + item["address"]).lower()
                or "example"
                in (item["email"] + " " + item["name"] + " " + item["address"]).lower()
            )


def test_no_search(client: TestClient):
    for e in engines:
        response = client.get(
            f"/api/v1/test/search_sample?limit=5&format=json&%24engine={e}"
        )
        assert response.status_code == 200
        jsd = response.json()
        assert len(jsd) >= 3
        assert len(jsd) <= 5
        assert "search_score" not in jsd[0]



import duckdb
from threading import active_count
from uuid import uuid4
import os
import psutil
import shutil
import deltalake

process = psutil.Process()


def naive_json_1(
    res: duckdb.DuckDBPyConnection, fn: str
):  # This works, no matter how often executed
    import json

    first = True
    with open(fn, "w") as f:
        f.write("[")
        with res.fetch_record_batch(1000) as reader:
            for batch in reader:
                aspy = batch.to_pylist()
                for i in aspy:
                    if first:
                        first = False
                    else:
                        f.write(", ")
                    f.write(json.dumps(i, default=str))
        f.write("]")


shutil.rmtree("out")
os.makedirs("out", exist_ok=True)
i = 0
while i < 100:
    print(f"{i} run")
    i += 1
    with duckdb.connect() as con:
        dt = deltalake.DeltaTable(
            "tests/data/delta/fake",
        )
        ds = dt.to_pyarrow_dataset()

        con.register("fake", ds)
        naive_json_1(con.execute("SELECT * FROM fake"), f"out/{str(uuid4())}.json")
        count = active_count()
        rss = process.memory_info().rss
        print(f"After Naive impl: NR Threads: {count}. RAM: {rss / 1000 / 1000} MB")
        con.execute(
            f"CREATE TEMP VIEW t1 AS SELECT * FROM fake order by 1; COPY (SELECT * FROM t1) TO 'out/{str(uuid4())}.json' (FORMAT JSON, Array True);drop VIEW t1"
        )  # Works only once
        count = active_count()
        rss = process.memory_info().rss
        print(f"NR Threads: {count}. RAM: {rss / 1000 / 1000} MB")



from typing import Optional, cast
import pyarrow as pa

import sqlglot.expressions as ex
from fastapi import (
    APIRouter,
    HTTPException,
    Query,
    Request,
)
from sqlglot import select

from bmsdna.lakeapi.context import get_context_by_engine
from bmsdna.lakeapi.core.config import BasicConfig, Config, Configs
from bmsdna.lakeapi.core.datasource import Datasource
from bmsdna.lakeapi.core.types import (
    Engines,
    MetadataDetailResult,
    MetadataSchemaField,
    MetadataSchemaFieldType,
)


def create_detailed_meta_endpoint(
    schema: Optional[pa.Schema],
    config: Config,
    configs: Configs,
    router: APIRouter,
    basic_config: BasicConfig,
):
    route = config.route + "/metadata_detail"
    has_complex = True
    if schema is not None:
        has_complex = any((pa.types.is_nested(t) for t in schema.types))

    @router.get(
        route,
        tags=["metadata", config.tag],
        operation_id=config.tag + "_" + config.name,
        name=config.name + "_metadata",
    )
    async def get_detailed_metadata(
        req: Request,
        jsonify_complex: bool = Query(
            title="jsonify_complex", include_in_schema=has_complex, default=False
        ),
        include_str_lengths: bool = True,
        engine: Engines | None = Query(
            title="$engine",
            alias="$engine",
            default=None,
            include_in_schema=False,
        ),
    ) -> MetadataDetailResult:
        req.state.lake_api_basic_config = basic_config
        if config.engine == "odbc":
            engine = None
        with get_context_by_engine(
            engine or config.engine or basic_config.default_engine,
            basic_config.default_chunk_size,
        ) as context:
            assert config.datasource is not None
            with Datasource(
                config.version_str,
                config.tag,
                config.name,
                config=config.datasource,
                sql_context=context,
                basic_config=basic_config,
                accounts=configs.accounts,
            ) as realdataframe:
                if not realdataframe.file_exists():
                    raise HTTPException(404)
                partition_columns = []
                partition_values = None
                delta_tbl = None
                df = realdataframe.get_df(None)
                if config.datasource.file_type == "delta":
                    delta_tbl = realdataframe.get_delta_table(schema_only=True)
                    assert delta_tbl is not None
                    assert delta_tbl.last_metadata is not None
                    partition_columns = delta_tbl.last_metadata.get(
                        "partitionColumns", []
                    )
                    partition_columns = [
                        c
                        for c in partition_columns
                        if not basic_config.should_hide_col_name(c)
                    ]  # also hide those from metadata detail
                    if len(partition_columns) > 0:
                        qb = cast(
                            ex.Select,
                            df.query_builder().select(
                                *[ex.column(c, quoted=True) for c in partition_columns],
                                append=False,
                            ),
                        ).distinct()
                        with context.execute_sql(qb) as res:
                            partition_values = await res.to_pylist()
                schema = df.arrow_schema()
                str_cols = [
                    name
                    for name in schema.names
                    if (
                        pa.types.is_string(schema.field(name).type)
                        or pa.types.is_large_string(schema.field(name).type)
                    )
                    and not basic_config.should_hide_col_name(name)
                ]
                complex_str_cols = (
                    [
                        name
                        for name in schema.names
                        if (
                            pa.types.is_struct(schema.field(name).type)
                            or pa.types.is_list(schema.field(name).type)
                        )
                        and not basic_config.should_hide_col_name(name)
                    ]
                    if jsonify_complex
                    else []
                )
                if include_str_lengths:
                    str_lengths_query = (
                        select(
                            *[
                                ex.func(
                                    "MAX",
                                    ex.func(
                                        context.len_func, ex.column(sc, quoted=True)
                                    ),
                                ).as_(sc)
                                for sc in str_cols + complex_str_cols
                            ]
                        )
                        .from_("strcols")
                        .with_(
                            "strcols",
                            as_=context.jsonify_complex(
                                df.query_builder(),
                                complex_str_cols,
                                str_cols + complex_str_cols,
                            ),
                        )
                    )
                    with context.execute_sql(str_lengths_query) as res:
                        str_lengths_df = (
                            (await res.to_pylist())
                            if len(str_cols) > 0 or len(complex_str_cols) > 0
                            else [{}]
                        )
                    str_lengths = str_lengths_df[0]
                else:
                    str_lengths = {}

                def _recursive_get_type(t: pa.DataType) -> MetadataSchemaFieldType:
                    is_complex = pa.types.is_nested(t)

                    return MetadataSchemaFieldType(
                        type_str=str(pa.string())
                        if is_complex and jsonify_complex
                        else str(t),
                        orig_type_str=str(t),
                        fields=(
                            [
                                MetadataSchemaField(
                                    name=f.name, type=_recursive_get_type(f.type)
                                )
                                for f in [
                                    t.field(find)
                                    for find in range(
                                        0, cast(pa.StructType, t).num_fields
                                    )
                                ]
                            ]
                            if pa.types.is_struct(t) and not jsonify_complex
                            else None
                        ),
                        inner=(
                            _recursive_get_type(cast(pa.ListType, t).value_type)
                            if pa.types.is_list(t)
                            or pa.types.is_large_list(t)
                            or pa.types.is_fixed_size_list(t)
                            and cast(pa.ListType, t).value_type is not None
                            and not jsonify_complex
                            else None
                        ),
                    )

                schema = df.arrow_schema()
                mdt = realdataframe.sql_context.get_modified_date(
                    realdataframe.source_uri, realdataframe.config.file_type
                )
                return MetadataDetailResult(
                    partition_values=partition_values,
                    partition_columns=partition_columns,
                    max_string_lengths=str_lengths,
                    data_schema=[
                        MetadataSchemaField(
                            name=n,
                            type=_recursive_get_type(schema.field(n).type),
                            max_str_length=str_lengths.get(n, None),
                        )
                        for n in schema.names
                        if not basic_config.should_hide_col_name(n)
                    ],
                    delta_meta=delta_tbl.last_metadata if delta_tbl else None,
                    delta_schema=delta_tbl.schema if delta_tbl else None,
                    parameters=config.params,  # type: ignore
                    search=config.search,
                    modified_date=mdt,
                )



from datetime import datetime

from fastapi.concurrency import run_in_threadpool
import pyarrow as pa
from typing import List, Optional, Tuple, Any, Union, cast
from bmsdna.lakeapi.core.types import FileTypes, OperatorType
from bmsdna.lakeapi.context.df_base import (
    FLAVORS,
    ExecutionContext,
    ResultData,
    get_sql,
)
import arrow_odbc
import pyarrow.dataset
import sqlglot.expressions as ex
import os
from bmsdna.lakeapi.core.config import SearchConfig
from sqlglot import from_, parse_one
from uuid import uuid4
from .source_uri import SourceUri

ENABLE_COPY_TO = os.environ.get("ENABLE_COPY_TO", "0") == "1"


def _get_temp_table_name():
    return "temp_" + str(uuid4()).replace("-", "")


arrow_odbc.enable_odbc_connection_pooling()


class BatchReaderWrap:
    def __init__(self, rdr: arrow_odbc.BatchReader):
        self.rdr = rdr

    def __enter__(self, *args, **kwargs):
        return self

    @property
    def schema(self):
        return self.rdr.schema

    def __iter__(self):
        return self.rdr.__iter__()

    def __exit__(self, *args, **kwargs):
        pass


class ODBCResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[ex.Query, str],
        connection_string: str,
        chunk_size: int,
    ) -> None:
        super().__init__(chunk_size=chunk_size)
        self.original_sql = original_sql
        self.connection_string = connection_string
        self._arrow_schema = None
        self._df = None
        self.flavor: FLAVORS = (
            "tsql" if " for SQL Server".lower() in connection_string.lower() else "ansi"
        )
        self.dialect = (
            "tsql"
            if " for SQL Server".lower() in connection_string.lower()
            else "duckdb"
        )

    def query_builder(self) -> ex.Select:
        if not isinstance(self.original_sql, str):
            return from_(self.original_sql.subquery().as_("t"))
        else:
            return from_(
                cast(ex.Select, parse_one(self.original_sql, dialect=self.dialect))
                .subquery()
                .as_("t")
            )

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = get_sql(self.original_sql, limit=0, dialect=self.dialect)
        batches = arrow_odbc.read_arrow_batches_from_odbc(
            query, connection_string=self.connection_string, batch_size=self.chunk_size
        )
        assert batches is not None
        self._arrow_schema = batches.schema
        return self._arrow_schema

    async def get_df(self):
        if self._df is None:
            query = get_sql(self.original_sql, dialect=self.dialect)
            batch_reader = await run_in_threadpool(
                arrow_odbc.read_arrow_batches_from_odbc,
                query,
                connection_string=self.connection_string,
                batch_size=self.chunk_size,
            )
            assert batch_reader is not None
            self._df = pa.Table.from_batches(batch_reader, batch_reader.schema)
        return self._df

    async def to_pandas(self):
        return (await self.get_df()).to_pandas()

    async def to_pylist(self):
        df = await self.get_df()
        return df.to_pylist()

    async def to_arrow_recordbatch(self, chunk_size: int = 10000):  # type: ignore
        query = get_sql(self.original_sql, dialect=self.dialect)
        res = await run_in_threadpool(
            arrow_odbc.read_arrow_batches_from_odbc,
            query,
            connection_string=self.connection_string,
            batch_size=self.chunk_size,
        )
        assert res is not None
        return BatchReaderWrap(res)


class ODBCExecutionContext(ExecutionContext):
    def __init__(self, chunk_size: int):
        super().__init__(chunk_size=chunk_size, engine_name="odbc")
        self.res_con = None
        self.datasources = dict()
        self.persistance_file_name = None

    def register_arrow(
        self, name: str, ds: Union[pyarrow.dataset.Dataset, pyarrow.Table]
    ):
        raise NotImplementedError("Cannot read arrow in remote sql")

    def close(self):
        pass

    @property
    def dialect(self):
        if len(self.datasources) > 0:
            return (
                "tsql"
                if " for SQL Server".lower()
                in self.datasources[[list(self.datasources.keys())[0]]]
                else "postgres"
            )
        return "tsql"

    @property
    def supports_view_creation(self) -> bool:
        return False

    def execute_sql(
        self,
        sql: Union[
            ex.Query,
            str,
        ],
    ) -> ODBCResultData:
        # todo: get correct connection string somehow
        assert len(self.datasources) == 1
        return ODBCResultData(
            sql,
            chunk_size=self.chunk_size,
            connection_string=self.datasources[list(self.datasources.keys())[0]],
        )

    def json_function(self, term: ex.Expression, assure_string=False):
        raise NotImplementedError(
            "Cannot convert to JSON in remote sql"
        )  # we could but sql does not support structured types anyway, so...

    def init_search(
        self,
        source_view: str,
        search_configs: list[SearchConfig],
    ):
        raise NotImplementedError("Not supported")

    def register_datasource(
        self,
        target_name: str,
        source_table_name: Optional[str],
        uri: SourceUri,
        file_type: FileTypes,
        filters: Any,
        meta_only: bool = False,
    ):
        assert file_type == "odbc"
        assert uri.account is None
        self.datasources[target_name] = uri.uri

    def list_tables(self) -> ResultData:
        return self.execute_sql(
            "SELECT table_schema, table_name as name, table_type from information_schema.tables"
        )

    def get_modified_date(
        self,
        uri: SourceUri,
        file_type: FileTypes,
    ) -> datetime | None:
        return None

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass



from typing import Optional, Union
from fastapi import APIRouter, BackgroundTasks, Header, Query, Request
from bmsdna.lakeapi.context.df_base import ExecutionContext, FileTypeNotSupportedError
from bmsdna.lakeapi.core.config import BasicConfig, Configs
from bmsdna.lakeapi.core.datasource import Datasource
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.types import OutputFileType
from bmsdna.lakeapi.core.response import create_response
from bmsdna.lakeapi.context import get_context_by_engine, Engines


sql_contexts: dict[str, ExecutionContext] = {}

logger = get_logger(__name__)


def init_duck_con(
    con: ExecutionContext,
    basic_config: BasicConfig,
    configs: Configs,
):
    for cfg in configs:
        assert cfg.datasource is not None
        with Datasource(
            cfg.version_str,
            cfg.tag,
            cfg.name,
            config=cfg.datasource,
            sql_context=con,
            accounts=configs.accounts,
            basic_config=basic_config,
        ) as df:
            if cfg.engine != "odbc" and df.file_exists():
                try:
                    con.register_datasource(
                        df.unique_table_name,
                        df.tablename,
                        df.execution_uri,
                        df.config.file_type,
                        None,
                    )
                except (FileTypeNotSupportedError, FileNotFoundError):
                    logger.warning(f"Cannot query {df.tablename}")


def _get_sql_context(
    engine: Engines,
    basic_config: BasicConfig,
    configs: Configs,
):
    assert engine not in ["odbc", "sqlite"]
    global sql_contexts
    if engine not in sql_contexts:
        sql_contexts[engine] = get_context_by_engine(
            engine,
            basic_config.default_chunk_size,
        )
        init_duck_con(
            sql_contexts[engine],
            basic_config,
            configs,
        )
        if basic_config.prepare_sql_db_hook is not None:
            basic_config.prepare_sql_db_hook(
                sql_contexts[engine],
            )

    return sql_contexts[engine]


def create_sql_endpoint(
    router: APIRouter,
    basic_config: BasicConfig,
    configs: Configs,
):
    @router.on_event("shutdown")
    async def shutdown_event():
        global sql_contexts
        for item in sql_contexts.values():
            item.__exit__()
        sql_contexts = {}

    @router.get("/api/sql/tables", tags=["sql"], operation_id="get_sql_tables")
    async def get_sql_tables(
        request: Request,
        background_tasks: BackgroundTasks,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        engine: Engines = Query(
            title="$engine", alias="$engine", default="duckdb", include_in_schema=False
        ),
    ):
        con = _get_sql_context(engine, basic_config, configs)
        return await con.list_tables().to_pylist()

    @router.post(
        "/api/sql",
        tags=["sql"],
        operation_id="post_sql_endpoint",
    )
    async def get_sql_post(
        request: Request,
        background_tasks: BackgroundTasks,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        engine: Engines = Query(
            title="$engine",
            alias="$engine",
            default="duckdb",
            include_in_schema=False,
        ),
    ):
        body = await request.body()

        con = _get_sql_context(engine, basic_config, configs)

        sql = body.decode("utf-8")

        return await create_response(
            request.url,
            request.query_params,
            format or request.headers["Accept"],
            con,
            sql,
            basic_config=basic_config,
            close_context=False,
        )

    @router.get(
        "/api/sql",
        tags=["sql"],
        operation_id="get_sql_endpoint",
    )
    async def get_sql_get(
        request: Request,
        background_tasks: BackgroundTasks,
        sql: str,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        engine: Engines = Query(
            title="$engine",
            alias="$engine",
            default="duckdb",
            include_in_schema=False,
        ),
    ):
        con = _get_sql_context(engine, basic_config, configs)

        return await create_response(
            request.url,
            request.query_params,
            format or request.headers["Accept"],
            con,
            sql,
            basic_config=basic_config,
            close_context=False,
        )



from abc import abstractmethod, ABC
from datetime import datetime

from sqlglot import Dialect
from bmsdna.lakeapi.core.types import FileTypes, OperatorType
from typing import Sequence, Literal, Optional, List, Tuple, Any, TYPE_CHECKING, Union
import pyarrow as pa
import sqlglot.expressions as ex

from bmsdna.lakeapi.utils.async_utils import _async
from deltalake2db import FilterType

from .source_uri import SourceUri

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow.dataset as pas
    from bmsdna.lakeapi.core.config import SearchConfig

FLAVORS = Literal["ansi", "tsql"]


def is_complex_type(
    schema: pa.Schema,
    col_name: str,
):
    import pyarrow.types as pat

    f = schema.field(col_name)
    return pat.is_nested(f.type)


def get_sql(
    sql_or_pypika: str | ex.Query,
    limit: int | None = None,
    *,
    dialect: str | Dialect,
) -> str:
    if not isinstance(sql_or_pypika, str) and dialect == "tsql":
        from_ = sql_or_pypika.args.get(
            "from"
        )  # for sql server always add an alias for sub queries
        if (
            from_
            and not from_.alias
            and isinstance(from_.args.get("this"), ex.Subquery)
            and not from_.args.get("this").alias
        ):
            sub_query = from_.args.get("this")
            sub_query.as_("s", copy=False)
    if limit is not None:
        sql_or_pypika = (
            sql_or_pypika.limit(limit)
            if not isinstance(sql_or_pypika, str)
            # why not just support limit/offset like everyone else, microsoft?
            else (
                f"SELECT * FROM ({sql_or_pypika}) s LIMIT {limit} "
                if dialect != "tsql"
                else f"SELECT top {limit} * FROM ({sql_or_pypika}) s "
            )
        )
    if isinstance(sql_or_pypika, str):
        return sql_or_pypika
    if len(sql_or_pypika.expressions) == 0:
        sql_or_pypika = sql_or_pypika.select("*")
    assert not isinstance(sql_or_pypika, str)
    return sql_or_pypika.sql(dialect=dialect)


class ResultData(ABC):
    def __init__(self, chunk_size: int) -> None:
        super().__init__()
        self.chunk_size = chunk_size

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def columns(self):
        return (self.arrow_schema()).names

    @abstractmethod
    def arrow_schema(self) -> "pa.Schema": ...

    @abstractmethod
    async def to_pylist(self) -> Sequence[dict]: ...

    @abstractmethod
    async def to_arrow_recordbatch(
        self, chunk_size: int = 10000
    ) -> "pa.RecordBatchReader": ...

    @abstractmethod
    def query_builder(self) -> ex.Select: ...

    async def write_json(self, file_name: str):
        import polars as pl

        with open(file_name, mode="wb") as f:
            pld = pl.from_arrow(await _async(self.to_arrow_recordbatch()))
            assert not isinstance(pld, pl.Series)
            pld.write_json(f)

    async def to_json(self):
        full_list = []
        for item in await _async(self.to_arrow_recordbatch()):
            full_list.extend(item.to_pylist())
        import pydantic

        return pydantic.TypeAdapter(list[dict]).dump_json(full_list)

    async def to_ndjson(self):
        import polars as pl

        result_strings = []
        for item in await _async(self.to_arrow_recordbatch()):
            res = pl.from_arrow(item)
            assert not isinstance(res, pl.Series)
            result_strings.append(res.write_ndjson())
        return "\n".join(result_strings)

    async def write_nd_json(self, file_name: str):
        import polars as pl

        batches = await self.to_arrow_recordbatch(self.chunk_size)
        with open(file_name, mode="wb") as f:
            for batch in batches:
                t = pl.from_arrow(batch)
                assert isinstance(t, pl.DataFrame)
                t.write_ndjson(f)

    @abstractmethod
    async def to_pandas(self) -> "pd.DataFrame": ...

    async def write_parquet(self, file_name: str):
        import pyarrow.parquet as paparquet

        batches = await self.to_arrow_recordbatch(self.chunk_size)

        with paparquet.ParquetWriter(
            file_name,
            batches.schema,
        ) as writer:
            for batch in batches:
                writer.write_batch(batch)

    async def write_csv(self, file_name: str, *, separator: str):
        import pyarrow.csv as pacsv

        batches = await self.to_arrow_recordbatch(self.chunk_size)
        with pacsv.CSVWriter(
            file_name,
            batches.schema,
            write_options=pacsv.WriteOptions(delimiter=separator),
        ) as writer:
            for batch in batches:
                writer.write_batch(batch)


class FileTypeNotSupportedError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class ExecutionContext(ABC):
    def __init__(self, chunk_size: int, engine_name: str) -> None:
        super().__init__()
        self.modified_dates: dict[str, datetime | None] = {}
        self.chunk_size = chunk_size
        self.len_func = "LEN"
        self.engine_name = engine_name

        self.array_contains_func = "array_contains"

    def term_like(
        self,
        a: ex.Expression,
        value: str,
        wildcard_loc: Literal["start", "end", "both"],
        *,
        negate=False,
    ):
        if wildcard_loc == "start":
            return a.like("%" + value) if not negate else ~a.like("%" + value)
        elif wildcard_loc == "end":
            return a.like(value + "%") if not negate else ~a.like(value + "%")
        else:
            return (
                a.like("%" + value + "%") if not negate else ~a.like("%" + value + "%")
            )

    @property
    @abstractmethod
    def dialect(self) -> str | Dialect: ...

    @property
    @abstractmethod
    def supports_view_creation(self) -> bool: ...

    @abstractmethod
    def __enter__(self) -> "ExecutionContext": ...

    @abstractmethod
    def __exit__(self, *args, **kwargs): ...

    def get_pyarrow_dataset(
        self,
        uri: SourceUri,
        file_type: FileTypes,
        filters: Optional[FilterType],
    ) -> "Optional[pas.Dataset | pa.Table]":
        spec_fs, spec_uri = uri.get_fs_spec()
        match file_type:
            case "parquet":
                import pyarrow.dataset as ds

                return ds.dataset(
                    spec_uri,
                    filesystem=spec_fs,
                    format=ds.ParquetFileFormat(
                        read_options=ds.ParquetReadOptions(
                            coerce_int96_timestamp_unit="us", dictionary_columns=None
                        ),
                    ),  # type: ignore
                )  # type: ignore
            case "ipc" | "arrow" | "feather" | "csv" | "orc":
                import pyarrow.dataset as ds

                return ds.dataset(
                    spec_uri,
                    filesystem=spec_fs,
                    format=file_type,
                )
            case "ndjson" | "json":
                import pandas
                import pyarrow

                ab_uri, ab_opts = uri.get_uri_options(flavor="fsspec")
                pd = pandas.read_json(
                    ab_uri,
                    storage_options=ab_opts,
                    orient="records",
                    lines=file_type == "ndjson",
                )

                return pyarrow.Table.from_pandas(pd)
            case "delta":
                from deltalake2db.delta_meta_retrieval import get_meta, PolarsEngine

                ab_uri, ab_opts = uri.get_uri_options(flavor="object_store")
                meta = get_meta(PolarsEngine(ab_opts), ab_uri)
                assert meta.protocol is not None
                if meta.protocol["minReaderVersion"] > 1:
                    raise ValueError(
                        "Delta table protocol version not supported, use DuckDB or Polars"
                    )
                files = meta.add_actions.keys()
                import pyarrow.dataset as ds

                return ds.dataset(files, filesystem=spec_fs, format="parquet")  # type: ignore
            case _:
                raise FileTypeNotSupportedError(
                    f"Not supported file type {file_type}",
                )

    @abstractmethod
    def register_arrow(
        self,
        name: str,
        ds: "Union[pas.Dataset, pa.Table]",
    ): ...

    @abstractmethod
    def close(self): ...

    def init_search(
        self,
        source_view: str,
        search_configs: "list[SearchConfig]",
    ):
        pass

    def init_spatial(self):
        pass

    @abstractmethod
    def json_function(
        self,
        term: ex.Expression,
        assure_string=False,
    ) -> ex.Expression: ...

    def jsonify_complex(
        self, query: ex.Query, complex_cols: list[str], columns: list[str]
    ) -> ex.Query:
        return query.select(
            *[
                ex.column(c, quoted=True)
                if c not in complex_cols
                else self.json_function(ex.column(c, quoted=True)).as_(c)
                for c in columns
            ]
        )

    def distance_m_function(
        self,
        lat1: ex.Expression,
        lon1: ex.Expression,
        lat2: ex.Expression,
        lon2: ex.Expression,
    ) -> ex.Expression:
        # haversine which works for duckdb and polars and probably most sql systems
        def acos(t):
            return ex.func("acos", t)

        def cos(t):
            return ex.func("cos", t)

        def radians(t):
            return ex.func("radians", t)

        def sin(t):
            return ex.func("sin", t)

        return ex.convert(6371000) * acos(
            cos(radians(lat1)) * cos(radians(lat2)) * cos(radians(lon2) - radians(lon1))
            + sin(radians(lat1)) * sin(radians(lat2))
        )

    def search_score_function(
        self,
        source_view: str,
        search_text: str,
        search_config: "SearchConfig",
        alias: Optional[str],
    ) -> ex.Expression:
        assert len(search_text) > 2
        parts = search_text.split(" ")

        cases = []
        summ = None
        for part in parts:
            case = ex.case()
            cond = ex.Concat(
                expressions=[ex.column(c) for c in search_config.columns]
            ).like("%" + part + "%")
            case.when(cond, ex.convert(1), copy=False)
            case.else_(ex.convert(0), copy=False)
            cases.append(case)
            summ = case if summ is None else summ + case
        assert summ is not None
        ni = ex.Nullif(this=summ, expression=ex.convert(0))
        return ni.as_(alias) if alias else ni

    def get_modified_date(
        self,
        uri: SourceUri,
        file_type: FileTypes,
    ) -> datetime | None:
        if file_type == "odbc":
            return None
        fs, fs_uri = uri.get_fs_spec()
        if not fs.exists(fs_uri):
            return None
        if file_type == "delta":
            try:
                from deltalake2db.delta_meta_retrieval import get_meta, PolarsEngine

                ab_uri, ab_opts = uri.get_uri_options(flavor="object_store")
                meta = get_meta(PolarsEngine(ab_opts), ab_uri)
                return meta.last_write_time
            except FileNotFoundError:
                return None

        fs, fs_uri = uri.get_fs_spec()
        return fs.modified(fs_uri)

    def register_datasource(
        self,
        target_name: str,
        source_table_name: Optional[str],
        uri: SourceUri,
        file_type: FileTypes,
        filters: Optional[FilterType],
        meta_only: bool = False,
    ):
        ds = self.get_pyarrow_dataset(uri, file_type, filters)
        self.modified_dates[target_name] = self.get_modified_date(uri, file_type)
        assert ds is not None
        self.register_arrow(target_name, ds)

    @abstractmethod
    def execute_sql(self, sql: Union[ex.Query, str]) -> ResultData: ...

    @abstractmethod
    def list_tables(self) -> ResultData: ...



import pytest
from fastapi.testclient import TestClient

engines = ["duckdb", "polars"]


@pytest.mark.parametrize("engine", engines)
def test_parquet(engine, client: TestClient):
    response = client.get(
        f"/api/v1/blobb/blob_test?format=json&limit=50&$engine={engine}",
    )
    assert response.status_code == 200
    fakedt = response.json()
    assert len(fakedt) == 50


@pytest.mark.parametrize("engine", engines)
def test_parquet_copied(engine, client: TestClient):
    response = client.get(
        f"/api/v1/blobb/blob_test_copied?format=json&limit=50&$engine={engine}",
    )
    assert response.status_code == 200
    fakedt = response.json()
    assert len(fakedt) == 50

    response2 = client.get(
        f"/api/v1/blobb/blob_test_copied?format=json&limit=50&$engine={engine}",
    )
    assert response2.status_code == 200
    fakedt2 = response2.json()
    assert len(fakedt2) == 50


@pytest.mark.parametrize("engine", engines)
def test_delta(engine, client: TestClient):
    response = client.get(
        f"/api/v1/blobb/fake?format=json&limit=50&$engine={engine}",
    )
    assert response.status_code == 200
    fakedt = response.json()
    assert len(fakedt) == 50



from threading import active_count
from uuid import uuid4
import os
import psutil
import shutil
import deltalake
import pyarrow.dataset as ds

process = psutil.Process()


def naive_json_1(res: ds.Dataset, fn: str):  # This works, no matter how often executed
    import json

    first = True
    with open(fn, "w") as f:
        f.write("[")
        for batch in res.to_batches(batch_size=1000):
            aspy = batch.to_pylist()
            for i in aspy:
                if first:
                    first = False
                else:
                    f.write(", ")
                f.write(json.dumps(i, default=str))
        f.write("]")


if os.path.exists("out"):
    shutil.rmtree("out")
os.makedirs("out", exist_ok=True)
i = 0
while i < 100:
    print(f"{i} run")
    i += 1

    dt = deltalake.DeltaTable(
        "tests/data/delta/fake",
    )
    dss = dt.to_pyarrow_dataset()

    count = active_count()
    rss = process.memory_info().rss
    print(f"before Naive impl: NR Threads: {count}. RAM: {rss / 1000 / 1000} MB")
    naive_json_1(dss, f"out/{str(uuid4())}.json")
